/*
 * wsp_session.c - Implement WSP session oriented service
 *
 * Lars Wirzenius
 */


#include <string.h>
#include <limits.h>

#include "gwlib/gwlib.h"
#include "wsp.h"
#include "wsp_pdu.h"
#include "wsp_headers.h"
#include "wsp_caps.h"
#include "wsp_strings.h"
#include "cookies.h"
#include "wap.h"
#include "wtp.h"


typedef enum {
#define STATE_NAME(name) name,
#define ROW(state, event, condition, action, next_state)
#include "wsp_client_session_states.def"

#define STATE_NAME(name) name,
#define ROW(state, event, condition, action, next_state)
#include "wsp_client_method_states.def"

#define STATE_NAME(name) name,
#define ROW(state, event, condition, action, next_state)
#include "wsp_client_push_states.def"

    WSPState_count
} WSPState;


/*
 * Give the status the module:
 *
 *	limbo
 *		not running at all
 *	running
 *		operating normally
 *	terminating
 *		waiting for operations to terminate, returning to limbo
 */
static enum {
    limbo, running, terminating
} run_status = limbo;

static wap_dispatch_func_t *dispatch_to_wtp_init;
static wap_dispatch_func_t *dispatch_to_wtp_init;
static wap_dispatch_func_t *dispatch_to_appl;

/*
 * True iff "Session resume facility" is enabled.  This means we are
 * willing to let sessions go to SUSPENDED state, and later resume them.
 * Currently we always support it, but this may become configurable
 * at some point.
 */

static int resume_enabled = 1;

static List *queue = NULL;
static List *session_machines = NULL;

static Counter *client_push_id_counter = NULL;


static WSPClientMachine *find_session_machine(WAPEvent *event);
static void handle_session_event(WSPClientMachine *sm, WAPEvent *current_event, WSP_PDU *pdu);
static void cant_handle_event(WSPClientMachine *sm, WAPEvent *event);
static WSPClientMachine *machine_create(void);
static void destroy_methodmachines(List *machines);
static void destroy_pushmachines(List *machines);
static void machine_destroy(void *pp);
static void handle_method_event(WSPClientMachine *sm, WSPClientMethodMachine *msm, WAPEvent *current_event, WSP_PDU *pdu);
static WSPClientMethodMachine *method_machine_create(WSPClientMachine *sm, long wtp_handle);
static void method_machine_destroy(void *p);
static WSPClientPushMachine *client_push_machine_create(WSPClientMachine *sm);
static void push_machine_destroy(void *p);
static char *state_name(WSPState state);
static Octstr *make_connect_pdu(WSPClientMachine *m);
static Octstr *make_disconnect_pdu(WSPClientMachine *m);
static Octstr *make_get_pdu(List *httpHeaders, Octstr *uri);
static Octstr *make_post_pdu(List *httpHeaders, Octstr *uri, Octstr *data);
static Octstr *make_suspend_pdu(WSPClientMachine *m);
static Octstr *make_resume_pdu(List *caps, List *headers);
static int transaction_belongs_to_session(void *wsp_ptr, void *tuple_ptr);
static int find_by_session_handle(void *wsp_ptr, void *id_ptr);
static int find_by_method_id(void *wspm_ptr, void *id_ptr);
static int find_by_push_id(void *m_ptr, void *id_ptr);
static WSPClientMethodMachine *find_client_method_machine(WSPClientMachine *sm, long id);
static WSPClientPushMachine *find_push_machine(WSPClientMachine *m, long id);
static int same_client(void *a, void *b);
static void disconnect_other_sessions(WSPClientMachine *sm);
static List *unpack_new_headers(WSPClientMachine *sm, Octstr *hdrs);
static WAPEvent *make_abort(long reason, long handle);
static void send_abort(long reason, long handle);
static void sendResultResponse(long handle);
static void sendResultDataResponse(long handle);
static void send_invoke(WSPClientMachine *m, Octstr *pdu, long handle, long acknowledgement, long class);
static void indicate_disconnect(WSPClientMachine *sm, long reason);
static void indicate_suspend(WSPClientMachine *sm, long reason);
static void indicate_resume(WSPClientMachine *sm, WAPAddrTuple *tuple, List *headers);
static void indicate_push(WSPClientMachine *sm, WSP_PDU *pdu);
static void indicate_pushabort(WSPClientPushMachine *spm, long reason);
static void indicate_confirmedpush(WSPClientMachine *sm, WSP_PDU *pdu, long client_push_id);
static void confirm_connect(WSPClientMachine *sm, List *httpHeaders, List *caps);
static void confirm_resume(WSPClientMachine *sm, List *httpHeaders);
static void confirm_methodinvoke(WSPClientMethodMachine *msm);
static void confirm_push(WSPClientPushMachine *m);
static void method_abort(WSPClientMethodMachine *msm, long reason);
static void indicate_method_result(WSPClientMethodMachine *msm, List *headers, Octstr *body, long status);
static void indicate_method_abort(WSPClientMethodMachine *msm, long reason);
static int method_is_holding(void *item, void *pattern);
static void release_holding_methods(WSPClientMachine *sm);
static void abort_methods(WSPClientMachine *sm, long reason);
static void abort_method(WSPClientMachine *sm, long handle, long reason);
static void abort_pushes(WSPClientMachine *sm, long reason);
static void handle_push_event(WSPClientMachine *sm, WSPClientPushMachine *pm, WAPEvent *current_event, WSP_PDU *pdu);

static void main_thread(void *);
static int id_belongs_to_session (void *, void *);



/***********************************************************************
 * Public functions.
 */


void wsp_client_session_init(wap_dispatch_func_t *initiator_dispatch, wap_dispatch_func_t *application_dispatch) {
    queue = list_create();
    list_add_producer(queue);
    session_machines = list_create();
	client_push_id_counter = counter_create();
    dispatch_to_wtp_init = initiator_dispatch;
    dispatch_to_appl = application_dispatch;
    wsp_strings_init();
    run_status = running;
    gwthread_create(main_thread, NULL);
}


void wsp_client_session_shutdown(void) {
    gw_assert(run_status == running);
    run_status = terminating;
    list_remove_producer(queue);
    counter_destroy(client_push_id_counter);
    gwthread_join_every(main_thread);

    list_destroy(queue, wap_event_destroy_item);

    debug("wap.wsp", 0, "WSP: %ld session machines left.", list_len(session_machines));
    list_destroy(session_machines, machine_destroy);

    wsp_strings_shutdown();
}


void wsp_client_session_dispatch_event(WAPEvent *event) {
    wap_event_assert(event);
    list_produce(queue, event);
}


/***********************************************************************
 * Local functions
 */


static void main_thread(void *arg) {
    WAPEvent *e;
    WSPClientMachine *sm;
    WSP_PDU *pdu;

    while (run_status == running && (e = list_consume(queue)) != NULL) {
        wap_event_assert(e);

        /* the content of the following WTP primitives sometimes needs to be inspected */
        switch (e->type) {
            case TR_Invoke_Ind:
                pdu = wsp_pdu_unpack(e->u.TR_Invoke_Ind.user_data);
                if (pdu == NULL) {
                    warning(0, "WSP: Broken PDU ignored.");
                    wap_event_destroy(e);
                    continue;
                }
                break;
            case TR_Result_Ind:
                pdu = wsp_pdu_unpack(e->u.TR_Result_Ind.user_data);
                if (pdu == NULL) {
                    warning(0, "WSP: Broken PDU ignored.");
                    wap_event_destroy(e);
                    continue;
                }
                break;
            default:
                pdu = NULL;
                break;
        }

        sm = find_session_machine(e);
        if (sm == NULL) {
            wap_event_destroy(e);
        } else {
            handle_session_event(sm, e, pdu);
        }

        wsp_pdu_destroy(pdu);
    }
}


static WSPClientMachine *find_session_machine(WAPEvent *event) {
    WSPClientMachine *sm;
    long session_handle;
    WAPAddrTuple *tuple;

    tuple = NULL;
    session_handle = -1;

    switch (event->type) {
        case TR_Invoke_Ind:
            tuple = wap_addr_tuple_duplicate(event->u.TR_Invoke_Ind.addr_tuple);
            break;
        case TR_Invoke_Cnf:
            tuple = wap_addr_tuple_duplicate(event->u.TR_Invoke_Cnf.addr_tuple);
            break;
        case TR_Result_Cnf:
            tuple = wap_addr_tuple_duplicate(event->u.TR_Result_Cnf.addr_tuple);
            break;
        case TR_Abort_Ind:
            tuple = wap_addr_tuple_duplicate(event->u.TR_Abort_Ind.addr_tuple);
            break;
            /* linking address tuple to WTP handle */
        case TR_Invoke_Req:
            tuple = wap_addr_tuple_duplicate(event->u.TR_Invoke_Req.addr_tuple);
            break;
        case TR_Result_Ind:
            tuple = wap_addr_tuple_duplicate(event->u.TR_Result_Ind.addr_tuple);
            break;

        case Disconnect_Event:
            session_handle = event->u.Disconnect_Event.session_handle;
            break;
        case Suspend_Event:
            session_handle = event->u.Suspend_Event.session_handle;
            break;


        case S_Connect_Req:
            tuple = wap_addr_tuple_duplicate(event->u.S_Connect_Req.addr_tuple);
            break;
        case S_Disconnect_Req:
            session_handle = event->u.S_Disconnect_Req.session_handle;
            break;
        case S_MethodInvoke_Req:
            session_handle = event->u.S_MethodInvoke_Req.session_handle;
            break;
        case S_MethodAbort_Req:
            session_handle = event->u.S_MethodAbort_Req.session_handle;
            break;
        case S_Connect_Res:
            session_handle = event->u.S_Connect_Res.session_id;
            break;
        case S_MethodResult_Res:
            session_handle = event->u.S_MethodResult_Res.session_handle;
            break;
        case S_Suspend_Req:
            session_handle = event->u.S_Suspend_Req.session_handle;
            break;
        case S_Resume_Req:
            session_handle = event->u.S_Resume_Req.session_handle;
            break;

        case S_ConfirmedPush_Res:
            session_handle = event->u.S_ConfirmedPush_Res.session_handle;
            break;
        case S_PushAbort_Req:
            session_handle = event->u.S_PushAbort_Req.session_handle;
            break;



        default:
            error(0, "WSP: Cannot find machine for %s event",
                  wap_event_name(event->type));
    }

    gw_assert(tuple != NULL || session_handle != -1);

    sm = NULL;
    if (session_handle != -1) {
        sm = list_search(session_machines, &session_handle, find_by_session_handle);
    } else {
        sm = list_search(session_machines, tuple, transaction_belongs_to_session);
    }

    if (sm != NULL && event->type == S_Connect_Req) {
        /* need to disonnect that session */
        disconnect_other_sessions(sm);
        sm = NULL;
    }

    if (sm == NULL && event->type == S_Connect_Req) {
        /* need a new state machine */
        sm = machine_create();
        sm->request_caps = wsp_cap_duplicate_list(event->u.S_Connect_Req.requested_capabilities);
        sm->http_headers = http_header_duplicate(event->u.S_Connect_Req.client_headers);
        sm->addr_tuple = tuple;
        sm->session_handle = event->u.S_Connect_Req.session_handle;
    } else
        wap_addr_tuple_destroy(tuple);

    return sm;
}


static void handle_session_event(WSPClientMachine *sm, WAPEvent *current_event, WSP_PDU *pdu) {
    
    debug("wap.wsp", 0, "WSP: machine %p, state %s, event %s", (void *) sm, state_name(sm->state), wap_event_name(current_event->type));

#define STATE_NAME(name)
#define ROW(state_name, event, condition, action, next_state) \
		{ \
			struct event *e; \
			e = &current_event->u.event; \
			if (sm->state == state_name && \
			   current_event->type == event && \
			   (condition)) { \
				action \
				sm->state = next_state; \
				debug("wap.wsp", 0, "WSP %ld: New state %s", \
					sm->session_handle, #next_state); \
				goto end; \
			} \
		}
#include "wsp_client_session_states.def"

    cant_handle_event(sm, current_event);

    end:
    wap_event_destroy(current_event);

    if (sm->state == NULL_SESSION)
        machine_destroy(sm);
}


static void cant_handle_event(WSPClientMachine *sm, WAPEvent *event) {
    /* We do the rest of the pre-state-machine tests here.  The first
     * four were done in find_session_machine().  The fifth is a
     * class 1 or 2 TR-Invoke.ind not handled by the state tables. */
    if (event->type == TR_Invoke_Ind &&
        (event->u.TR_Invoke_Ind.tcl == 1 ||
         event->u.TR_Invoke_Ind.tcl == 2)) {
        warning(0, "WSP: Can't handle TR-Invoke.ind, aborting transaction.");
        debug("wap.wsp", 0, "WSP: The unhandled event:");
        wap_event_dump(event);
        send_abort(WSP_ABORT_PROTOERR,
                   event->u.TR_Invoke_Ind.handle);
        /* The sixth is a class 0 TR-Invoke.ind not handled by state tables. */
    } else if (event->type == TR_Invoke_Ind) {
        warning(0, "WSP: Can't handle TR-Invoke.ind, ignoring.");
        debug("wap.wsp", 0, "WSP: The ignored event:");
        wap_event_dump(event);
        /* The seventh is any other event not handled by state tables. */
    } else {
        error(0, "WSP: Can't handle event. Aborting session.");
        debug("wap.wsp", 0, "WSP: The unhandled event:");
        wap_event_dump(event);
        /* TR-Abort.req(PROTOERR) if it is some other transaction
         * event than abort. */
        /* Currently that means TR-Result.cnf, because we already
         * tested for Invoke. */
        /* FIXME We need a better way to get at event values than
         * by hardcoding the types. */
        if (event->type == TR_Result_Cnf) {
            send_abort(WSP_ABORT_PROTOERR,
                       event->u.TR_Result_Cnf.handle);
        }
        /* Abort(PROTOERR) all method and push transactions */
        abort_methods(sm, WSP_ABORT_PROTOERR);
        abort_pushes(sm, WSP_ABORT_PROTOERR);
        /* S-Disconnect.ind(PROTOERR) */
        indicate_disconnect(sm, WSP_ABORT_PROTOERR);
    }
}


static WSPClientMachine *machine_create(void) {
    WSPClientMachine *p;

    p = gw_malloc(sizeof(WSPClientMachine));
    debug("wap.wsp", 0, "WSP: Created WSPClientMachine %p", (void *) p);

#define INTEGER(name) p->name = 0;
#define OCTSTR(name) p->name = NULL;
#define HTTPHEADERS(name) p->name = NULL;
#define ADDRTUPLE(name) p->name = NULL;
#define MACHINESLIST(name) p->name = list_create();
#define CAPABILITIES(name) p->name = NULL;
#define COOKIES(name) p->name = NULL;
#define REFERER(name) p->name = NULL;
#define MACHINE(fields) fields
#include "wsp_client_session_machine.def"

    p->state = NULL_SESSION;

    /* set capabilities to default values (defined in 1.1) */

    p->SDU_size = 1400;

    /* Insert new machine at the _front_, because 1) it's more likely
     * to get events than old machines are, so this speeds up the linear
     * search, and 2) we want the newest machine to get any method
     * invokes that come through before the Connect is established. */
    list_insert(session_machines, 0, p);

    /* handle */
    p->connect_handle = -1;
    p->resume_handle = -1;

    return p;
}


static void destroy_methodmachines(List *machines) {
    if (list_len(machines) > 0) {
        warning(0, "Destroying WSP session with %ld active methods\n",
                list_len(machines));
    }

    list_destroy(machines, method_machine_destroy);
}

static void destroy_pushmachines(List *machines) {
    if (list_len(machines) > 0) {
        warning(0, "Destroying WSP session with %ld active pushes\n",
                list_len(machines));
    }

    list_destroy(machines, push_machine_destroy);
}

static void machine_destroy(void *pp) {
    WSPClientMachine *p;

    p = pp;
    debug("wap.wsp", 0, "Destroying WSPClientMachine %p", pp);
    list_delete_equal(session_machines, p);

#define INTEGER(name) p->name = 0;
#define OCTSTR(name) octstr_destroy(p->name);
#define HTTPHEADERS(name) http_destroy_headers(p->name);
#define ADDRTUPLE(name) wap_addr_tuple_destroy(p->name);
#define MACHINESLIST(name) destroy_##name(p->name);
#define CAPABILITIES(name) wsp_cap_destroy_list(p->name);
#define COOKIES(name) cookies_destroy(p->name);
#define REFERER(name) octstr_destroy(p->name);
#define MACHINE(fields) fields
#include "wsp_client_session_machine.def"
    gw_free(p);
}


struct msm_pattern {
    WAPAddrTuple *addr_tuple;
    long msmid, tid;
};


/* This function does NOT consume its event; it leaves that task up
 * to the parent session */
static void handle_method_event(WSPClientMachine *sm, WSPClientMethodMachine *msm, WAPEvent *current_event, WSP_PDU *pdu) {

    if (msm == NULL) {
        warning(0, "No method machine for event.");
        wap_event_dump(current_event);
        return;
    }

    debug("wap.wsp", 0, "WSP: method %ld, state %s, event %s",
          msm->transaction_id, state_name(msm->state), 
          wap_event_name(current_event->type));

    gw_assert(sm->session_handle == msm->session_handle);

#define STATE_NAME(name)
#define ROW(state_name, event, condition, action, next_state) \
		{ \
			struct event *e; \
			e = &current_event->u.event; \
			if (msm->state == state_name && \
			   current_event->type == event && \
			   (condition)) { \
				action \
				msm->state = next_state; \
				debug("wap.wsp", 0, "WSP %ld/%ld: New method state %s", \
					msm->session_handle, msm->transaction_id, #next_state); \
				goto end; \
			} \
		}
#include "wsp_client_method_states.def"

    cant_handle_event(sm, current_event);

    end:
    if (msm->state == NULL_METHOD) {
        method_machine_destroy(msm);
        list_delete_equal(sm->methodmachines, msm);
    }
}


static WSPClientMethodMachine *method_machine_create(WSPClientMachine *sm, long wtp_handle) {
    WSPClientMethodMachine *msm;

    msm = gw_malloc(sizeof(*msm));

#define INTEGER(name) msm->name = 0;
#define ADDRTUPLE(name) msm->name = NULL;
#define EVENT(name) msm->name = NULL;
#define MACHINE(fields) fields
#include "wsp_client_method_machine.def"

    msm->transaction_id = wtp_handle;
    msm->state = NULL_METHOD;
    msm->addr_tuple = wap_addr_tuple_duplicate(sm->addr_tuple);
    msm->server_session_id = sm->server_session_id;
    msm->session_handle = sm->session_handle;

    list_append(sm->methodmachines, msm);

    return msm;
}



static void method_machine_destroy(void *p) {
    WSPClientMethodMachine *msm;

    if (p == NULL)
        return;

    msm = p;

    debug("wap.wsp", 0, "Destroying WSPClientMethodMachine %ld",
          msm->transaction_id);

#define INTEGER(name)
#define ADDRTUPLE(name) wap_addr_tuple_destroy(msm->name);
#define EVENT(name) wap_event_destroy(msm->name);
#define MACHINE(fields) fields
#include "wsp_client_method_machine.def"

    gw_free(msm);
}

static void handle_push_event(WSPClientMachine *sm, WSPClientPushMachine *pm, WAPEvent *current_event, WSP_PDU *pdu) {
    if (pm == NULL) {
        warning(0, "No push machine for event.");
        wap_event_dump(current_event);
        return;
    }

    debug("wap.wsp", 0, "WSP(tid/pid): push %ld/%ld, state %s, event %s",
          pm->transaction_id, pm->client_push_id, state_name(pm->state), 
          wap_event_name(current_event->type));
    gw_assert(sm->server_session_id == pm->server_session_id);

#define STATE_NAME(name)
#define ROW(state_name, event, condition, action, next_state) \
		{ \
		    if (pm->state == state_name && \
			current_event->type == event && \
			(condition)) { \
			     action \
			     pm->state = next_state; \
			     debug("wap.wsp", 0, "WSP %ld/%ld: New push state %s", \
			           pm->session_handle, pm->transaction_id, #next_state); \
				goto end; \
			} \
		}
#include "wsp_client_push_states.def"

    cant_handle_event(sm, current_event);
    end:
    if (pm->state == CLIENT_PUSH_NULL_STATE) {
        push_machine_destroy(pm);
        list_delete_equal(sm->pushmachines, pm);
    }
}

static WSPClientPushMachine *client_push_machine_create(WSPClientMachine *sm) {
    WSPClientPushMachine *m;

    m = gw_malloc(sizeof(WSPClientPushMachine));

#define INTEGER(name) m->name = 0;
#define ADDRTUPLE(name) m->name = NULL;
#define HTTPHEADER(name) m->name = http_create_empty_headers();
#define MACHINE(fields) fields
#include "wsp_client_push_machine.def"

    m->client_push_id = counter_increase(client_push_id_counter);
    m->state = CLIENT_PUSH_NULL_STATE;
    m->addr_tuple = wap_addr_tuple_duplicate(sm->addr_tuple);
    m->server_session_id = sm->server_session_id;
    m->session_handle = sm->session_handle;

    list_append(sm->pushmachines, m);

    return m;      
}

static void push_machine_destroy(void *p) {
    WSPClientPushMachine *m = NULL;   

    if (p == NULL)
        return;
    m = p;
    debug("wap.wsp", 0, "Destroying WSPClientPushMachine %ld",
          m->transaction_id);
#define INTEGER(name) 
#define ADDRTUPLE(name) wap_addr_tuple_destroy(m->name);
#define HTTPHEADER(name) http_destroy_headers(m->name);
#define MACHINE(fields) fields
#include "wsp_client_push_machine.def"

    gw_free(m);
}

static char *state_name(WSPState state) {
    switch (state) {
#define STATE_NAME(name) case name: return #name;
#define ROW(state, event, cond, stmt, next_state)
#include "wsp_client_session_states.def"

#define STATE_NAME(name) case name: return #name;
#define ROW(state, event, cond, stmt, next_state)
#include "wsp_client_method_states.def"

#define STATE_NAME(name) case name: return #name;
#define ROW(state, event, cond, stmt, next_state)
#include "wsp_client_push_states.def"

        default:
            return "unknown wsp state";
    }
}






static Octstr *make_connect_pdu(WSPClientMachine *m) {
    WSP_PDU *pdu;
    Octstr *os;

    pdu = wsp_pdu_create(Connect);

    pdu->u.Connect.version = 0x10;
    pdu->u.Connect.capabilities = wsp_cap_pack_list(m->request_caps);    
    pdu->u.Connect.headers = wsp_headers_pack(m->http_headers, 0);

    os = wsp_pdu_pack(pdu);
    wsp_pdu_destroy(pdu);

    return os;
}

static Octstr *make_disconnect_pdu(WSPClientMachine *m) {
    WSP_PDU *pdu;
    Octstr *os;

    pdu = wsp_pdu_create(Disconnect);
    pdu->u.Disconnect.sessionid = m->server_session_id;
    os = wsp_pdu_pack(pdu);
    wsp_pdu_destroy(pdu);
    return os;
}


static Octstr *make_get_pdu(List *httpHeaders, Octstr *uri) {
    WSP_PDU *pdu;
    Octstr *os;

    pdu = wsp_pdu_create(Get);
    pdu->u.Get.uri = octstr_duplicate(uri);
    pdu->u.Get.headers = wsp_headers_pack(httpHeaders, 0);

    os = wsp_pdu_pack(pdu);
    wsp_pdu_destroy(pdu);
    return os;
}


static Octstr *make_post_pdu(List *httpHeaders, Octstr *uri, Octstr *data) {
    WSP_PDU *pdu;
    Octstr *os;

    pdu = wsp_pdu_create(Post);
    pdu->u.Post.uri = octstr_duplicate(uri);
    /* assume that Content-Type is one of the headers */
    pdu->u.Post.headers = wsp_headers_pack(httpHeaders, 1);
    pdu->u.Post.data = octstr_duplicate(data);

    os = wsp_pdu_pack(pdu);
    wsp_pdu_destroy(pdu);
    return os;
}

static Octstr *make_suspend_pdu(WSPClientMachine *m) {
    WSP_PDU *pdu;
    Octstr *os;

    pdu = wsp_pdu_create(Suspend);
    pdu->u.Suspend.sessionid = m->server_session_id;
    os = wsp_pdu_pack(pdu);
    wsp_pdu_destroy(pdu);
    return os;
}


static Octstr *make_resume_pdu(List *caps, List *headers) {
    WSP_PDU *pdu;
    Octstr *os;

    pdu = wsp_pdu_create(Resume);
    pdu->u.Resume.capabilities = wsp_cap_pack_list(caps);    
    pdu->u.Resume.headers = wsp_headers_pack(headers, 0);

    os = wsp_pdu_pack(pdu);
    wsp_pdu_destroy(pdu);

    return os;
}



static int transaction_belongs_to_session(void *wsp_ptr, void *tuple_ptr) {
    WSPClientMachine *wsp;
    WAPAddrTuple *tuple;

    wsp = wsp_ptr;
    tuple = tuple_ptr;

    return wap_addr_tuple_same(wsp->addr_tuple, tuple);
}


static int find_by_session_handle(void *wsp_ptr, void *id_ptr) {
    WSPClientMachine *wsp = wsp_ptr;
    long *idp = id_ptr;

    return wsp->session_handle == *idp;
}


static int find_by_method_id(void *wspm_ptr, void *id_ptr) {
    WSPClientMethodMachine *msm = wspm_ptr;
    long *idp = id_ptr;

    return msm->transaction_id == *idp;
}

static int find_by_push_id(void *m_ptr, void *id_ptr) {
    WSPClientPushMachine *m = m_ptr;
    long *idp = id_ptr;

    return m->transaction_id == *idp;
}

static WSPClientMethodMachine *find_client_method_machine(WSPClientMachine *sm, long id) {
    return list_search(sm->methodmachines, &id, find_by_method_id);
}

static WSPClientPushMachine *find_push_machine(WSPClientMachine *m, long id) {
    return list_search(m->pushmachines, &id, find_by_push_id);
}

static int same_client(void *a, void *b) {
    WSPClientMachine *sm1, *sm2;

    sm1 = a;
    sm2 = b;
    return wap_addr_tuple_same(sm1->addr_tuple, sm2->addr_tuple);
}


static void disconnect_other_sessions(WSPClientMachine *sm) {
    List *old_sessions;
    WAPEvent *disconnect;
    WSPClientMachine *sm2;
    long i;

    old_sessions = list_search_all(session_machines, sm, same_client);
    if (old_sessions == NULL)
        return;

    for (i = 0; i < list_len(old_sessions); i++) {
        sm2 = list_get(old_sessions, i);
        if (sm2 != sm) {
            disconnect = wap_event_create(Disconnect_Event);
            handle_session_event(sm2, disconnect, NULL);
        }
    }

    list_destroy(old_sessions, NULL);
}


static List *unpack_new_headers(WSPClientMachine *sm, Octstr *hdrs) {
    List *new_headers;

    if (hdrs && octstr_len(hdrs) > 0) {
        new_headers = wsp_headers_unpack(hdrs, 0);
        if (sm->http_headers == NULL)
            sm->http_headers = http_create_empty_headers();
        http_header_combine(sm->http_headers, new_headers);
        return new_headers;
    }
    return NULL;
}

static WAPEvent *make_abort(long reason, long handle) {
    WAPEvent *wtp_event;

    wtp_event = wap_event_create(TR_Abort_Req);
    wtp_event->u.TR_Abort_Req.abort_type = 0x01;
    wtp_event->u.TR_Abort_Req.abort_reason = reason;
    wtp_event->u.TR_Abort_Req.handle = handle;

    return wtp_event;
}

static void send_abort(long reason, long handle) {
    WAPEvent *wtp_event;

    wtp_event = make_abort(reason, handle);
    dispatch_to_wtp_init(wtp_event);
}

static void sendResultResponse(long handle) {
    WAPEvent *wtp_event;

    /* TR-Result.res */
    wtp_event = wap_event_create(TR_Result_Res);
    wtp_event->u.TR_Result_Res.handle = handle;
    dispatch_to_wtp_init(wtp_event);
}

static void sendResultDataResponse(long handle) {
    WAPEvent *wtp_event;

    /* TR-Result.res */
    wtp_event = wap_event_create(TR_Result_Res);
    wtp_event->u.TR_Result_Res.handle = handle;
    dispatch_to_wtp_init(wtp_event);
}


static void send_invoke(WSPClientMachine *m, Octstr *pdu, long handle, long acknowledgement, long class) {
    WAPEvent *wtp_event;

    wtp_event = wap_event_create(TR_Invoke_Req);
    wtp_event->u.TR_Invoke_Req.addr_tuple = wap_addr_tuple_duplicate(m->addr_tuple);
    wtp_event->u.TR_Invoke_Req.up_flag = acknowledgement;
    wtp_event->u.TR_Invoke_Req.tcl = class;
    wtp_event->u.TR_Invoke_Req.handle = handle;
    wtp_event->u.TR_Invoke_Req.user_data = pdu;

    dispatch_to_wtp_init(wtp_event);
}

static void indicate_disconnect(WSPClientMachine *sm, long reason) {
    WAPEvent *new_event;

    new_event = wap_event_create(S_Disconnect_Ind);
    new_event->u.S_Disconnect_Ind.reason_code = reason;
    new_event->u.S_Disconnect_Ind.redirect_security = 0;
    new_event->u.S_Disconnect_Ind.redirect_addresses = 0;
    new_event->u.S_Disconnect_Ind.error_headers = NULL;
    new_event->u.S_Disconnect_Ind.error_body = NULL;
    new_event->u.S_Disconnect_Ind.session_handle = sm->session_handle;
    dispatch_to_appl(new_event);
}



static void indicate_suspend(WSPClientMachine *sm, long reason) {
    WAPEvent *new_event;

    new_event = wap_event_create(S_Suspend_Ind);
    new_event->u.S_Suspend_Ind.reason = reason;
    new_event->u.S_Suspend_Ind.session_id = sm->server_session_id;
    dispatch_to_appl(new_event);
}


static void indicate_resume(WSPClientMachine *sm, WAPAddrTuple *tuple, List *headers) {
    WAPEvent *new_event;

    new_event = wap_event_create(S_Resume_Ind);
    new_event->u.S_Resume_Ind.addr_tuple = wap_addr_tuple_duplicate(tuple);
    new_event->u.S_Resume_Ind.client_headers = http_header_duplicate(headers);
    new_event->u.S_Resume_Ind.session_id = sm->session_handle;
    dispatch_to_appl(new_event);
}

static void indicate_push(WSPClientMachine *sm, WSP_PDU *pdu) {
    WAPEvent *new_event;

    new_event = wap_event_create(S_Push_Ind);
    new_event->u.S_Push_Ind.push_headers = wsp_headers_unpack(pdu->u.Push.headers, 1);
    new_event->u.S_Push_Ind.push_body = octstr_duplicate(pdu->u.Push.data);
    new_event->u.S_Push_Ind.session_handle = sm->session_handle;
    dispatch_to_appl(new_event);
}

static void indicate_pushabort(WSPClientPushMachine *spm, long reason) {
    WAPEvent *ota_event;

    ota_event = wap_event_create(S_PushAbort_Ind);
    ota_event->u.S_PushAbort_Ind.push_id = spm->client_push_id;
    ota_event->u.S_PushAbort_Ind.reason = reason;
    ota_event->u.S_PushAbort_Ind.session_id = spm->session_handle;
    dispatch_to_appl(ota_event);
}

static void indicate_confirmedpush(WSPClientMachine *sm, WSP_PDU *pdu, long client_push_id) {
    WAPEvent *new_event;

    new_event = wap_event_create(S_ConfirmedPush_Ind);
    new_event->u.S_ConfirmedPush_Ind.client_push_id = client_push_id;
    new_event->u.S_ConfirmedPush_Ind.push_headers = wsp_headers_unpack(pdu->u.ConfirmedPush.headers, 1);
    new_event->u.S_ConfirmedPush_Ind.push_body = octstr_duplicate(pdu->u.ConfirmedPush.data);
    new_event->u.S_ConfirmedPush_Ind.session_handle = sm->session_handle;

}

static void confirm_connect(WSPClientMachine *sm, List *httpHeaders, List *caps) {
    WAPEvent *new_event;

    new_event = wap_event_create(S_Connect_Cnf);
    new_event->u.S_Connect_Cnf.negotiated_capabilities = wsp_cap_duplicate_list(caps);
    new_event->u.S_Connect_Cnf.server_headers = http_header_duplicate(httpHeaders);
    new_event->u.S_Connect_Cnf.session_handle = sm->session_handle;
    dispatch_to_appl(new_event);
}

static void confirm_resume(WSPClientMachine *sm, List *httpHeaders) {
    WAPEvent *new_event;

    new_event = wap_event_create(S_Resume_Cnf);
    new_event->u.S_Resume_Cnf.server_headers = http_header_duplicate(httpHeaders);
    new_event->u.S_Resume_Cnf.session_handle = sm->session_handle;
    dispatch_to_appl(new_event);
}

static void confirm_methodinvoke(WSPClientMethodMachine *msm) {
    WAPEvent *new_event;

    new_event = wap_event_create(S_MethodInvoke_Cnf);
    new_event->u.S_MethodInvoke_Cnf.client_transaction_id = msm->transaction_id;
    new_event->u.S_MethodInvoke_Cnf.session_handle = msm->session_handle;
    dispatch_to_appl(new_event);
}


static void method_abort(WSPClientMethodMachine *msm, long reason) {
    WAPEvent *wtp_event;

    /* Send TR-Abort.req(reason) */
    wtp_event = wap_event_create(TR_Abort_Req);
    /* FIXME: Specs are unclear about this; we may indeed have to
     * guess abort whether this is a WSP or WTP level abort code */
    if (reason < WSP_ABORT_PROTOERR) {
        wtp_event->u.TR_Abort_Req.abort_type = 0x00;
    } else {
        wtp_event->u.TR_Abort_Req.abort_type = 0x01;
    }
    wtp_event->u.TR_Abort_Req.abort_reason = reason;
    wtp_event->u.TR_Abort_Req.handle = msm->transaction_id;

    dispatch_to_wtp_init(wtp_event);
}

static void indicate_method_result(WSPClientMethodMachine *msm, List *headers, Octstr *body, long status) {
    WAPEvent *new_event;

    /* Send S-MethodResult.ind */
    new_event = wap_event_create(S_MethodResult_Ind);
    new_event->u.S_MethodResult_Ind.client_transaction_id = msm->transaction_id;
    new_event->u.S_MethodResult_Ind.status = status;
    new_event->u.S_MethodResult_Ind.response_headers = http_header_duplicate(headers);
    new_event->u.S_MethodResult_Ind.response_body = octstr_duplicate(body);
    new_event->u.S_MethodResult_Ind.session_handle = msm->session_handle;
    dispatch_to_appl(new_event);
}


static void indicate_method_abort(WSPClientMethodMachine *msm, long reason) {
    WAPEvent *new_event;

    /* Send S-MethodAbort.ind(reason) */
    new_event = wap_event_create(S_MethodAbort_Ind);
    new_event->u.S_MethodAbort_Ind.transaction_id = msm->transaction_id;
    new_event->u.S_MethodAbort_Ind.reason = reason;
    new_event->u.S_MethodAbort_Ind.session_handle = msm->session_handle;
    dispatch_to_appl(new_event);
}


static void abort_methods(WSPClientMachine *sm, long reason) {
    WAPEvent *ab;
    WSPClientMethodMachine *msm;
    long i, len;

    ab = wap_event_create(Abort_Event);
    ab->u.Abort_Event.reason = reason;

    /* This loop goes backward because it has to deal with the
     * possibility of method machines disappearing after their event. */
    len = list_len(sm->methodmachines);
    for (i = len - 1; i >= 0; i--) {
        msm = list_get(sm->methodmachines, i);
        handle_method_event(sm, msm, ab, NULL);
    }

    wap_event_destroy(ab);
}

static void abort_method(WSPClientMachine *sm, long handle, long reason) {
    WSPClientMethodMachine *msm;
    WAPEvent *ab;

    /* abort(PROTOERR) the method transaction */        
    ab = wap_event_create(Abort_Event);
    ab->u.Abort_Event.reason = reason;
    msm = find_client_method_machine(sm, handle);
    handle_method_event(sm, msm, ab, NULL);
}

static void abort_pushes(WSPClientMachine *sm, long reason) {
    WAPEvent *ab;
    WSPClientPushMachine *psm;
    long i, len;

    ab = wap_event_create(Abort_Event);
    ab->u.Abort_Event.reason = reason;

    len = list_len(sm->pushmachines);
    for (i = len - 1; i >= 0; i--) {
        psm = list_get(sm->pushmachines, i);
        handle_push_event(sm, psm, ab, NULL);
    }

    wap_event_destroy(ab);
}










